/*
 *  CreateRTFFromMLTE.cpp
 *  MLTERTFLib
 *
 *  Created by James Walker on 7/2/06.
 *
 *  Copyright (c) 2006 James W. Walker.
 *
 *	This software is provided 'as-is', without any express or implied warranty.
 *	In no event will the authors be held liable for any damages arising from the
 *	use of this software.
 *
 *	Permission is granted to anyone to use this software for any purpose,
 *	including commercial applications, and to alter it and redistribute it
 *	freely, subject to the following restrictions:
 *
 *	1. The origin of this software must not be misrepresented; you must not
 *	claim that you wrote the original software. If you use this software in a
 *	product, an acknowledgment in the product documentation would be appreciated
 *	but is not required.
 *
 *	2. Altered source versions must be plainly marked as such, and must not be
 *	misrepresented as being the original software.
 *
 *	3. This notice may not be removed or altered from any source distribution.
 */

#include "CreateRTFFromMLTE.h"

#include "CFRefHolder.h"

#include <vector>
#include <sstream>
#include <algorithm>
#include <string>
#include <cmath>
#include <iomanip>

extern const ATSUAttributeTag	kATSULinkURLHack;

namespace
{
	const RGBColor	kBlackColor = { 0, 0, 0 };
	
	enum EBaselineShift
	{
		kBaselineShift_None = 0,
		kBaselineShift_Super,
		kBaselineShift_Sub
	};

	#pragma mark struct MLTEStyleRun
	struct MLTEStyleRun
	{
							MLTEStyleRun();
							MLTEStyleRun( const MLTEStyleRun& inOther );
							MLTEStyleRun(
									CFStringRefHolder inText,
									CFStringRefHolder inFont,
									Style inStyle,
									float inSize,
									const RGBColor& inColor,
									EBaselineShift inShift,
									bool inStrikethrough,
									CFURLRef inLinkURL );
							
							~MLTEStyleRun() {}
		
		MLTEStyleRun&		operator=( const MLTEStyleRun& inOther );
						
		CFStringRefHolder	mText;
		CFStringRefHolder	mFontName;
		Style				mQDStyle;
		float				mFontSize;
		RGBColor			mColor;
		EBaselineShift		mShift;
		bool				mStrikethrough;
		CFURLRefHolder		mLinkURL;
	};
	
	typedef std::vector< MLTEStyleRun >	RunVec;
	
	typedef	std::vector< CFStringRefHolder >	NameVec;
	
	typedef	std::vector< RGBColor >				ColorVec;
	
	struct NameEqualPredicate
	{
		bool	operator()( const CFStringRefHolder& inStr ) const
				{
					CFComparisonResult compRes = CFStringCompare( inStr.get(),
						mName, 0 );
					return compRes == kCFCompareEqualTo;
				}
				
		CFStringRef	mName;
	};
	
	struct ColorEqualPredicate
	{
		bool	operator()( const RGBColor& inColor ) const
				{
					return (inColor.red == mColor.red) and
						(inColor.green == mColor.green) and
						(inColor.blue == mColor.blue);
				}
		
		RGBColor	mColor;
	};
}


MLTEStyleRun::MLTEStyleRun()
	: mQDStyle( normal )
	, mFontSize( 12 )
	, mColor( kBlackColor )
	, mShift( kBaselineShift_None )
	, mStrikethrough( false )
{
}

MLTEStyleRun::MLTEStyleRun( const MLTEStyleRun& inOther )
	: mText( inOther.mText )
	, mFontName( inOther.mFontName )
	, mQDStyle( inOther.mQDStyle )
	, mFontSize( inOther.mFontSize )
	, mColor( inOther.mColor )
	, mShift( inOther.mShift )
	, mStrikethrough( inOther.mStrikethrough )
	, mLinkURL( inOther.mLinkURL )
{
}

MLTEStyleRun::MLTEStyleRun(
								CFStringRefHolder inText,
								CFStringRefHolder inFont,
								Style inStyle,
								float inSize,
								const RGBColor& inColor,
								EBaselineShift inShift,
								bool inStrikethrough,
								CFURLRef inLinkURL )
	: mText( inText )
	, mFontName( inFont )
	, mQDStyle( inStyle )
	, mFontSize( inSize )
	, mColor( inColor )
	, mShift( inShift )
	, mStrikethrough( inStrikethrough )
	, mLinkURL( inLinkURL )
{
	if (mLinkURL.get() != NULL)
	{
		CFRetain( mLinkURL.get() );
	}
}

MLTEStyleRun&	MLTEStyleRun::operator=( const MLTEStyleRun& inOther )
{
	mText = inOther.mText;
	mFontName = inOther.mFontName;
	mQDStyle = inOther.mQDStyle;
	mFontSize = inOther.mFontSize;
	mColor = inOther.mColor;
	mShift = inOther.mShift;
	mStrikethrough = inOther.mStrikethrough;
	mLinkURL = inOther.mLinkURL;
	
	return *this;
}

/*!
	@function	CopyStringFromRun
	@abstract	Create a CFStringRef containing the Unicode text
				of an MLTE style run.
	@param		inTextOb	An MLTE text object.
	@param		inRunStart	Starting offset of the run.
	@param		inRunEnd	Ending offset of the run.
	@result		The run text.
*/
static CFStringRef CopyStringFromRun(
						TXNObject inTextOb,
						TXNOffset inRunStart,
						TXNOffset inRunEnd )
{
	CFStringRef	theResult = NULL;
	
	Handle	dataH = NULL;
	OSStatus	err = TXNGetData( inTextOb, inRunStart, inRunEnd, &dataH );
	if (err == noErr)
	{
		Size	dataSize = GetHandleSize( dataH );
		ItemCount	numChars = dataSize / sizeof(UniChar);
		
		theResult = CFStringCreateWithCharacters( NULL, (const UniChar*)*dataH,
			numChars );
		
		DisposeHandle( dataH );
	}
	
	return theResult;
}

/*!
	@function	CopyFontNameFromID
	@abstract	Convert an ATSUFontID to a PostScript font name.
	@param		inFontID	An ATSU font ID number.
	@result		A reference to a PostScript font name.
*/
static CFStringRef	CopyFontNameFromID( ATSUFontID inFontID )
{
	CFStringRef	fontName = NULL;
	ATSFontRef	theFontRef = FMGetATSFontRefFromFont( inFontID );
	ATSFontGetPostScriptName( theFontRef, kATSOptionFlagsDefault,
		&fontName );
	return fontName;
}


/*!
	@function	MLTEToRuns
	@abstract	Split the text into style runs.
	@param		inTextOb		An MLTE text object.
	@param		inStartOffset	Starting offset for text to be converted to RTF.
	@param		inEndOffset		Ending offset for text to be converted to RTF.
	@param		outRuns			Receives the style runs.
*/
static void MLTEToRuns( TXNObject inTextOb,
						TXNOffset inStartOffset,
						TXNOffset inEndOffset,
						RunVec& outRuns )
{
	ItemCount	runCount, run;
	TXNCountRunsInRange( inTextOb, inStartOffset, inEndOffset, &runCount );

	TXNOffset	runStart, runEnd;
	TXNDataType	runType;
	RGBColor	qdColor;
	Style		qdStyle;
	Fixed		fontSize;
	ATSUFontID	fontID;
	Fixed		theShift;
	Boolean		theStrikethrough;
	
	TXNTypeAttributes	runAtt = {
		kTXNATSUIStyle, kTXNATSUIStyleSize
	};

	for (run = 0; run < runCount; ++run)
	{
		OSStatus err = TXNGetIndexedRunInfoFromRange( inTextOb, run,
			inStartOffset, inEndOffset,
			&runStart, &runEnd, &runType,
			1, &runAtt );
		
		if ( (err == noErr) && (runType == kTXNUnicodeTextData) )
		{
			ATSUStyle	theStyle = static_cast<ATSUStyle>(
				runAtt.data.dataPtr );
			runStart = std::max( runStart, inStartOffset );
			runEnd = std::min( runEnd, inEndOffset );
			
			// Extract attributes from the ATSUStyle.
			err = ATSUGetAttribute( theStyle, kATSUFontTag, sizeof(ATSUFontID),
				&fontID, NULL );
			err = ATSUGetAttribute( theStyle, kATSUSizeTag, sizeof(Fixed),
				&fontSize, NULL );
			err = ATSUGetAttribute( theStyle, kATSUColorTag, sizeof(RGBColor),
				&qdColor, NULL );
			if (err != noErr)
			{
				qdColor = kBlackColor;
			}
			err = ATSUGetAttribute( theStyle, kATSUCrossStreamShiftTag,
				sizeof(Fixed), &theShift, NULL );
			if (err != noErr)
			{
				theShift = 0;
			}
			err = ATSUGetAttribute( theStyle, kATSUStyleStrikeThroughTag,
				sizeof(Boolean), &theStrikethrough, NULL );
			if (err != noErr)
			{
				theStrikethrough = false;
			}
			CFURLRef	theURL = NULL;
			err = ATSUGetAttribute( theStyle, kATSULinkURLHack,
				sizeof(CFURLRef), &theURL, NULL );
			Boolean	qdItalic = false;
			err = ATSUGetAttribute( theStyle, kATSUQDItalicTag,
				sizeof(Boolean), &qdItalic, NULL );
			Boolean	qdBold = false;
			err = ATSUGetAttribute( theStyle, kATSUQDBoldfaceTag,
				sizeof(Boolean), &qdBold, NULL );
			Boolean	qdUnderline = false;
			err = ATSUGetAttribute( theStyle, kATSUQDUnderlineTag,
				sizeof(Boolean), &qdUnderline, NULL );
			qdStyle = (qdBold? bold : normal) |
				(qdItalic? italic : normal) |
				(qdUnderline? underline : normal);
			
			// Construct a style run.
			CFStringRefHolder	runText( CopyStringFromRun( inTextOb,
				runStart, runEnd ) );
			CFStringRefHolder runFont( CopyFontNameFromID( fontID ) );
			EBaselineShift	shiftCode = kBaselineShift_None;
			if (theShift > 0)
			{
				shiftCode = kBaselineShift_Super;
			}
			else if (theShift < 0)
			{
				shiftCode = kBaselineShift_Sub;
			}
			
			if ( (runText.get() != NULL) && (runFont.get() != NULL) )
			{
				MLTEStyleRun	theRun( runText, runFont, qdStyle,
					fontSize/65536.0f, qdColor, shiftCode, theStrikethrough,
					theURL );
				outRuns.push_back( theRun );
			}
		}
	}
}

/*!
	@function	FontNameToIndex
	@abstract	Find a font name in a table of names.
	@param		inName		A font name.
	@param		inTable		A vector of names.
	@result		A 0-based index, or -1 if not found.
*/
static int	FontNameToIndex( CFStringRef inName, const NameVec& inTable )
{
	int	theIndex = -1;
	
	NameEqualPredicate	tester = { inName };
	NameVec::const_iterator	foundIt = std::find_if( inTable.begin(),
		inTable.end(), tester );
	if (foundIt != inTable.end())
	{
		theIndex = (foundIt - inTable.begin());
	}
	
	return theIndex;
}

/*!
	@function	MakeFontTableForRuns
	@abstract	Make a table of the font names appearing in a vector of style
				runs.
	@param		inRuns		A vector of style runs.
	@param		outNames	Receives a vector of names.
*/
static void MakeFontTableForRuns( const RunVec& inRuns, NameVec& outNames )
{
	for (RunVec::const_iterator i = inRuns.begin(); i != inRuns.end(); ++i)
	{
		int	theIndex = FontNameToIndex( i->mFontName.get(), outNames );
		
		if (theIndex < 0)
		{
			outNames.push_back( i->mFontName );
		}
	}
}

/*!
	@function	ColorToIndex
	@abstract	Find the index of a color in a vector of colors.
	@param		inColor		A color.
	@param		inTable		A vector of colors.
	@result		A 0-based index, or -1 if not found.
*/
static int ColorToIndex( const RGBColor& inColor, const ColorVec& inTable )
{
	int	theIndex = -1;
	
	ColorEqualPredicate	tester = { inColor };
	ColorVec::const_iterator	foundIt = std::find_if( inTable.begin(),
		inTable.end(), tester );
	if (foundIt != inTable.end())
	{
		theIndex = (foundIt - inTable.begin());
	}
	
	return theIndex;
}

/*!
	@function	MakeColorTableForRuns
	@abstract	Construct the color table, given the complete list of style
				runs.
	@param		inRuns			The vector of style runs.
	@param		outColors		Receives the color table.
*/
static void MakeColorTableForRuns( const RunVec& inRuns, ColorVec& outColors )
{
	for (RunVec::const_iterator i = inRuns.begin(); i != inRuns.end(); ++i)
	{
		int	theIndex = ColorToIndex( i->mColor, outColors );
		
		if (theIndex < 0)
		{
			outColors.push_back( i->mColor );
		}
	}
}

/*!
	@function	FontNameToASCII
	@abstract	Convert a font name CFStringRef to an ASCII string.
	@discussion	I have never seen a PostScript font name that was not plain
				ASCII, but just in case I get the bytes with a loss byte.
	@param		inName		A PostScript font name.
	@result		The font name as an ASCII string.
*/
static std::string	FontNameToASCII( CFStringRef inName )
{
	CFIndex	nameLen = CFStringGetLength( inName );
	CFIndex	maxBytes = CFStringGetMaximumSizeForEncoding( nameLen,
		kCFStringEncodingASCII );
	std::vector<UInt8>	buffer( maxBytes + 1 );
	CFIndex	actualBytes;
	CFStringGetBytes( inName, CFRangeMake(0,nameLen), kCFStringEncodingASCII,
		'?', false, &buffer[0], maxBytes, &actualBytes );
	buffer[ actualBytes ] = 0;
	std::string	theName( reinterpret_cast<char*>( &buffer[0] ) );
	return theName;
}

/*!
	@function	FontNameToType
	
	@abstract	Choose a font family control word, recognizing certain common
				fonts.
	
	@param		inName		ASCII-encoded font name.
	@result		A font family control word with terminating space.
*/
static std::string	FontNameToType( const std::string inName )
{
	std::string	theType( "\\fnil " );
	
	if ( (inName.find("Times") != std::string::npos) or
		(inName.find("Palatino") != std::string::npos) or
		(inName.find("NewYork") != std::string::npos))
	{
		theType.assign( "\\froman " );
	}
	else if (inName.find("Courier") != std::string::npos)
	{
		theType.assign( "\\fmodern " );
	}
	else if ( (inName.find("Helvetica") != std::string::npos) or
		(inName.find("Arial") != std::string::npos) or
		(inName.find("Geneva") != std::string::npos) )
	{
		theType.assign( "\\fswiss " );
	}
	return theType;
}

/*!
	@function	FontTableToStream
	@abstract	Write the font table to the RTF stream.
	@discussion	Each entry in the font table looks like:
				\fN\fFAMILYTYPE FONTNAME;
	@param		inNames		Table of font names.
	@param		ioStream	The output stream.
*/
static void FontTableToStream( const NameVec& inNames, std::ostream& ioStream )
{
	ioStream << "{\\fonttbl";
	
	if (inNames.empty())
	{
		ioStream << "\\f0\\froman Times-Roman;";
	}
	else
	{
		for (int i = 0; i < inNames.size(); ++i)
		{
			std::string	theName( FontNameToASCII( inNames[i].get() ) );
			ioStream << "\\f" << i << FontNameToType( theName ) <<
				theName << ";";
		}
	}
	
	ioStream << "}\r";
}

/*!
	@function	ColorTableToStream
	@abstract	Write the color table to the RTF stream.
	@param		inColors	Table of colors.
	@param		ioStream	The output stream.
*/
static void ColorTableToStream( const ColorVec& inColors,
								std::ostream& ioStream )
{
	ioStream << "{\\colortbl";
	
	if (inColors.empty())
	{
		ioStream << "\\red0\\green0\\blue0;";
	}
	else
	{
		for (int i = 0; i < inColors.size(); ++i)
		{
			const RGBColor&	theColor( inColors[i] );
			
			ioStream << "\\red" << (theColor.red >> 8) <<
				"\\green" << (theColor.green >> 8) <<
				"\\blue" << (theColor.blue >> 8) << ";";
		}
	}
	
	ioStream << "}\r";
}

/*!
	@function	UpdateFont
	@abstract	If the font index has changed, write a font control word to the
				RTF stream.
	@param		inNewFontIndex		The new font index.
	@param		ioPrevFontIndex		The previous font index.
	@param		ioStream			The RTF output stream.
*/
static void UpdateFont( int inNewFontIndex, int& ioPrevFontIndex,
						std::ostream& ioStream )
{
	if (inNewFontIndex != ioPrevFontIndex)
	{
		ioStream << "\\f" << inNewFontIndex << " ";
		
		ioPrevFontIndex = inNewFontIndex;
	}
}

/*!
	@function	UpdateColor
	@abstract	If the color index has changed, write a color control word to
				the RTF stream.
	@param		inNewColorIndex		The new color index.
	@param		ioPrevColorIndex	The previous color index.
	@param		ioStream			The RTF output stream.
*/
static void UpdateColor( int inNewColorIndex, int& ioPrevColorIndex,
						std::ostream& ioStream )
{
	if (inNewColorIndex != ioPrevColorIndex)
	{
		ioStream << "\\cf" << inNewColorIndex << " ";
		
		ioPrevColorIndex = inNewColorIndex;
	}
}

/*!
	@function	UpdateSize
	@abstract	If the font size has changed, write a size control word to
				the RTF stream.
	@param		inNewSize			The new color index.
	@param		ioPrevSize			The previous color index.
	@param		ioStream			The RTF output stream.
*/
static void UpdateSize( int inNewSize, int& ioPrevSize,
						std::ostream& ioStream )
{
	if (inNewSize != ioPrevSize)
	{
		ioStream << "\\fs" << inNewSize << " ";
		
		ioPrevSize = inNewSize;
	}
}

/*!
	@function	UpdateShift
	@abstract	If the baseline shift has changed, write a shift control word to
				the RTF stream.
	@param		inNewShift			The new shift code.
	@param		ioPrevShift			The previous shift code.
	@param		ioStream			The RTF output stream.
*/
static void UpdateShift( EBaselineShift inNewShift,
						EBaselineShift& ioPrevShift,
						std::ostream& ioStream )
{
	if (inNewShift != ioPrevShift)
	{
		switch (inNewShift)
		{
			case kBaselineShift_None:
				ioStream << "\\nosupersub ";
				break;
			
			case kBaselineShift_Super:
				ioStream << "\\super ";
				break;
				
			case kBaselineShift_Sub:
				ioStream << "\\sub ";
				break;
		}
		ioPrevShift = inNewShift;
	}
}

/*!
	@function	UpdateStrike
	@abstract	If the strikethrough state has changed, write a strikethrough
				control word to the RTF stream.
	@param		inNewStrike			The new strikethrough state.
	@param		ioPrevStrike		The previous strikethrough state.
	@param		ioStream			The RTF output stream.
*/
static void UpdateStrike( Boolean inNewStrike, Boolean& ioPrevStrike,
						std::ostream& ioStream )
{
	if (inNewStrike != ioPrevStrike)
	{
		if (inNewStrike)
		{
			ioStream << "\\strike ";
		}
		else
		{
			ioStream << "\\strike0 ";
		}
		ioPrevStrike = inNewStrike;
	}
}

/*!
	@function	UpdateOneStyle
	@abstract	If two style differ in one specific bit, write the appropriate
				control word.
	@param		inNewStyle			New style.
	@param		inPrevStyle			Previous style.
	@param		inTestStyle			Style bit to test.
	@param		inControlWordText	Control word (minus the backslash and the
									parameter) for the test style.
	@param		ioStream			The RTF output stream.
*/
static void UpdateOneStyle( Style inNewStyle, Style inPrevStyle,
							Style inTestStyle, const char* inControlWordText,
							std::ostream& ioStream )
{
	if ( (inNewStyle & inTestStyle) != (inPrevStyle & inTestStyle) )
	{
		ioStream << '\\' << inControlWordText;
		if ( (inNewStyle & inTestStyle) == 0 )
		{
			ioStream << '0';
		}
	}
}

/*!
	@function	UpdateStyle
	@abstract	If the QuickDraw style has changed, write a style control word
				to the RTF stream.
	@param		inNewStyle			The new style.
	@param		ioPrevStyle			The previous style.
	@param		ioStream			The RTF output stream.
*/
static void UpdateStyle( Style inNewStyle, Style& ioPrevStyle,
						std::ostream& ioStream )
{
	UpdateOneStyle( inNewStyle, ioPrevStyle, bold, "b", ioStream );
	UpdateOneStyle( inNewStyle, ioPrevStyle, italic, "i", ioStream );
	UpdateOneStyle( inNewStyle, ioPrevStyle, underline, "ul", ioStream );
	
	if (inNewStyle != ioPrevStyle)
	{
		ioStream << ' ';
		
		ioPrevStyle = inNewStyle;
	}
}

/*!
	@function	RomanTextToRTFStream
	@abstract	Write some MacRoman text to an RTF stream.
	@discussion	Certain characters must be replaced by RTF control symbols and
				control words: backslash, curly braces, line breaks, and
				preferably curly quotes.  Any other non-ASCII characters must
				be hex-escaped.
	@param		inText		NUL-terminated MacRoman string.
	@param		ioStream	Stream of RTF code.
*/
static void RomanTextToRTFStream( const char* inText, std::ostream& ioStream )
{
	for (const char* p = inText; *p != 0; ++p)
	{
		UInt8	theChar = static_cast<UInt8>( *p );
		switch (theChar)
		{
			case '\\':
			case '{':
			case '}':
				ioStream << '\\' << static_cast<char>(theChar);
				break;
			
			case '\r':
			case '\n':
				ioStream << "\\par\r";
				break;
			
			case 0xD2:	// left double curly
				ioStream << "\\ldblquote ";
				break;
			
			case 0xD3:	// right double curly
				ioStream << "\\rdblquote ";
				break;
			
			case 0xD4:	// left single curly
				ioStream << "\\lquote ";
				break;
			
			case 0xD5:	// right single curly
				ioStream << "\\rquote ";
				break;
			
			case 0xA5:	// bullet
				ioStream << "\\bullet ";
				break;
			
			case 0xD1:	// en dash
				ioStream << "\\endash ";
				break;
			
			case 0xD0:	// em dash
				ioStream << "\\emdash ";
				break;
			
			default:
				if ((theChar & 0x80) == 0)	// ASCII
				{
					ioStream << static_cast<char>(theChar);
				}
				else
				{
					ioStream << "\\'" << std::hex << std::setfill('0') <<
						std::setw(2) << static_cast<int>(theChar) << std::dec;
				}
				break;
		}
	}
}

/*!
	@function	UnicodeTextToRTFStream
	@abstract	Write some Unicode text to an RTF stream.
	@discussion	Certain characters must be replaced by RTF control symbols and
				control words: backslash, curly braces, line breaks, and
				preferably curly quotes.  Any other non-ASCII characters must
				be escaped.
	@param		inText		Unicode string.
	@param		inNumChars	Number of characters.
	@param		ioStream	Stream of RTF code.
*/
static void UnicodeTextToRTFStream( const UniChar* inText, CFIndex inNumChars,
									std::ostream& ioStream )
{
	for (CFIndex i = 0; i < inNumChars; ++i)
	{
		UniChar	theChar = inText[ i ];
		
		switch (theChar)
		{
			case '\\':
			case '{':
			case '}':
				ioStream << '\\' << static_cast<char>(theChar);
				break;
			
			case '\r':
			case '\n':
				ioStream << "\\par\r";
				break;
			
			case 0x201C:	// left double curly
				ioStream << "\\ldblquote ";
				break;
			
			case 0x201D:	// right double curly
				ioStream << "\\rdblquote ";
				break;
			
			case 0x2018:	// left single curly
				ioStream << "\\lquote ";
				break;
			
			case 0x2019:	// right single curly
				ioStream << "\\rquote ";
				break;
			
			case 0x2022:	// bullet
				ioStream << "\\bullet ";
				break;
			
			case 0x2013:	// en dash
				ioStream << "\\endash ";
				break;
			
			case 0x2014:	// em dash
				ioStream << "\\emdash ";
				break;

			default:
				if (theChar < 0x80)	// ASCII
				{
					ioStream << static_cast<char>(theChar);
				}
				else
				{
					SInt16	signedChar = static_cast<SInt16>( theChar );
					ioStream << "\\u" << signedChar << '?';
				}
				break;
		}
	}
}

/*!
	@function	TextToRTFStream
	@abstract	Write some RTF-encoded text to the stream.
	@param		inText		Some text.
	@param		ioStream	The RTF output stream.
*/
static void TextToRTFStream( CFStringRef inText, std::ostream& ioStream )
{
	const CFIndex kNumChars = CFStringGetLength( inText );
	
	if (kNumChars > 0)
	{
		// First, see if we can convert to MacRoman.
		CFIndex	maxBytes = CFStringGetMaximumSizeForEncoding( kNumChars,
			kCFStringEncodingMacRoman );
		std::vector<char>	romanChars( maxBytes+1 );
		if (CFStringGetCString( inText, &romanChars[0],
			romanChars.size(), kCFStringEncodingMacRoman ))
		{
			RomanTextToRTFStream( &romanChars[0], ioStream );
		}
		else	// not all Roman, go for Unicode.
		{
			std::vector<UniChar>	theChars( kNumChars );
			CFStringGetCharacters( inText, CFRangeMake(0, kNumChars),
				&theChars[0] );
			UnicodeTextToRTFStream( &theChars[0], kNumChars, ioStream );
		}
	}
}

/*!
	@function	BeginLink
	@abstract	Write the RTF data that begins a URL link.
	@param		inURL		A URL.
	@param		ioStream	An output stream.
*/
static void BeginLink( CFURLRef inURL, std::ostream& ioStream )
{
	CFStringRef	urlStr = CFURLGetString( inURL );
	CFStringRefHolder	withoutEscapes(
		CFURLCreateStringByReplacingPercentEscapesUsingEncoding( NULL,
		urlStr, CFSTR(""), kCFStringEncodingUTF8 ) );
	CFStringRefHolder	withEscapes(
		CFURLCreateStringByAddingPercentEscapes( NULL, withoutEscapes.get(),
		NULL, NULL, kCFStringEncodingUTF8 ) );
	ioStream << "{\\field{\\*\\fldinst{HYPERLINK \"";
	TextToRTFStream( withEscapes.get(), ioStream );
	ioStream << "\"}}{\\fldrslt ";
}

/*!
	@function	RunsToRTFStream
	@abstract	Writes RTF data to a stream.
	@param		inRuns		A vector of style runs.
	@param		ioStream	An output stream.
*/
static void RunsToRTFStream( const RunVec& inRuns, std::ostream& ioStream )
{
	ioStream << "{\\rtf1\\mac\\ansicpg10000\r";
	
	NameVec	fontTable;
	MakeFontTableForRuns( inRuns, fontTable );
	FontTableToStream( fontTable, ioStream );
	
	ColorVec	colorTable;
	MakeColorTableForRuns( inRuns, colorTable );
	ColorTableToStream( colorTable, ioStream );
	
	int	prevFontIndex = -1;
	int	prevColorIndex = -1;
	Style prevStyle = normal;
	int	prevSize = -1;
	EBaselineShift prevShift = kBaselineShift_None;
	Boolean	prevStrike = false;
	
	for (RunVec::const_iterator i = inRuns.begin(); i != inRuns.end(); ++i)
	{
		const MLTEStyleRun&	theRun( *i );
		int	curFontIndex = FontNameToIndex( theRun.mFontName.get(), fontTable );
		UpdateFont( curFontIndex, prevFontIndex, ioStream );
		int curColorIndex = ColorToIndex( theRun.mColor, colorTable );
		UpdateColor( curColorIndex, prevColorIndex, ioStream );
		Style	curStyle = theRun.mQDStyle;
		UpdateStyle( curStyle, prevStyle, ioStream );
		int		curSize = lround( theRun.mFontSize * 2 );
		UpdateSize( curSize, prevSize, ioStream );
		UpdateShift( theRun.mShift, prevShift, ioStream );
		UpdateStrike( theRun.mStrikethrough, prevStrike, ioStream );
		if (theRun.mLinkURL.get() != NULL)
		{
			BeginLink( theRun.mLinkURL.get(), ioStream );
		}
		TextToRTFStream( theRun.mText.get(), ioStream );
		if (theRun.mLinkURL.get() != NULL)
		{
			ioStream << "}}";
		}
	}
	
	ioStream << "}\r";
}


/*!
	@function	RunsToRTF
	@abstract	Convert style runs to RTF data.
	@param		inRuns		A vector of style runs.
	@result		A data object containing RTF text.
*/
static CFDataRef RunsToRTF( const RunVec& inRuns )
{
	std::ostringstream	ss;
	
	RunsToRTFStream( inRuns, ss );
	
	std::string	streamData( ss.str() );
	
	CFDataRef	theData = CFDataCreate( NULL,
		reinterpret_cast<const UInt8*>( streamData.data() ),
		streamData.size() );
	
	return theData;
}

/*!
	@function	CreateRTFFromMLTE
	
	@abstract	Create RTF data encoding styled text in an MLTE object.
	
	@param		inTextOb	An MLTE object.
	@param		inStartOffset	Starting offset for text to be converted to RTF.
	@param		inEndOffset		Ending offset for text to be converted to RTF.
	
	@result		Data containing RTF code.
*/
CFDataRef	CreateRTFFromMLTE( TXNObject inTextOb,
								TXNOffset inStartOffset,
								TXNOffset inEndOffset )
{
	RunVec	theRuns;
	
	MLTEToRuns( inTextOb, inStartOffset, inEndOffset, theRuns );
	CFDataRef	theData = RunsToRTF( theRuns );
	return theData;
}
