/*
 *  LoadMLTEFromRTF.cpp
 *  MLTERTFLib
 *
 *  Created by James Walker on 7/3/06.
 *
 *  Copyright (c) 2006 James W. Walker.
 *
 *	This software is provided 'as-is', without any express or implied warranty.
 *	In no event will the authors be held liable for any damages arising from the
 *	use of this software.
 *
 *	Permission is granted to anyone to use this software for any purpose,
 *	including commercial applications, and to alter it and redistribute it
 *	freely, subject to the following restrictions:
 *
 *	1. The origin of this software must not be misrepresented; you must not
 *	claim that you wrote the original software. If you use this software in a
 *	product, an acknowledgment in the product documentation would be appreciated
 *	but is not required.
 *
 *	2. Altered source versions must be plainly marked as such, and must not be
 *	misrepresented as being the original software.
 *
 *	3. This notice may not be removed or altered from any source distribution.
 */

#include "LoadMLTEFromRTF.h"

#include "CFRefHolder.h"
#include "RTFDataStream.h"

#include <string>
#include <vector>
#include <map>

extern const ATSUAttributeTag	kATSULinkURLHack	= 'URL!';

namespace
{
	enum ERTFDestination
	{
		kRTFDestination_Main = 0,
		kRTFDestination_FontTable,
		kRTFDestination_ColorTable,
		kRTFDestination_StyleSheets,
		kRTFDestination_FieldInstructions,
		kRTFDestination_FieldResult,
		kRTFDestination_Unknown
	};

	typedef		std::vector< std::string >		StringVec;
	
	enum EFontType
	{
		kFontType_Unknown = 0,
		kFontType_Roman,
		kFontType_Modern,
		kFontType_Swiss
	};
	
	struct	FontEntry
	{
		std::string	mName;
		EFontType	mType;
	};
	
	typedef		std::map< int, FontEntry >	FontTable;
	
	struct CharPair
	{
		char	macChar;
		char	winChar;
	};
	
	typedef		std::map< std::string, CharPair >	TokenToChar;
	
	struct RTFColor
	{
		UInt8	red;
		UInt8	green;
		UInt8	blue;
	};
	
	const RTFColor	kBlackColor = { 0, 0, 0 };
	
	typedef		std::vector< RTFColor >			ColorVec;
	
	bool		operator==( const RTFColor& inOne, const RTFColor& inTwo );
	
	enum EBaselineShift
	{
		kBaselineShift_None = 0,
		kBaselineShift_Super,
		kBaselineShift_Sub
	};

	struct RTFStyle
	{
		int					mFontNumber;
		int					mColorIndex;
		int					mFontSizeHalfPoints;
		Style				mStyle;
		EBaselineShift		mShift;
		Boolean				mStrikethrough;
	};
	
	struct RTFStyleRun
	{
		CFStringRefHolder	mText;
		RTFStyle			mAtts;
		CFStringRefHolder	mLinkURL;
	};
	
	typedef		std::vector< RTFStyleRun >		RunVec;
	
	struct	RTFStateFrame
	{
		ERTFDestination		mDestination;
		int					mUnicodeAltBytes;
		RTFStyle			mAtts;
	};

	typedef		std::vector< RTFStateFrame >	StateVec;
	
	class CRTFParser
	{
	public:
							CRTFParser( CFDataRef inData );
							~CRTFParser();
		
		bool				ParseText();
		void				LoadText( TXNObject ioTextOb,
									TXNOffset inStartOffset,
									TXNOffset inEndOffset );
		
	private:
		typedef void (CRTFParser::*ControlSymbolHandler)( RTFParam inParam );
		
		struct CSDef
		{
			const char*				token;
			ControlSymbolHandler	handler;
		};
		
		typedef std::map< std::string, ControlSymbolHandler >
							HandlerMap;
		
		void				DoBeginGroup();
		void				DoEndGroup();
		void				DoControlWord( const std::string& inToken,
											RTFParam inParam );
		void				DoControlSymbol( char inToken,
											RTFParam inParam );
		void				DoPlainCharacter( char inChar );
		void				SetDestination( ERTFDestination inDest );
		void				LookForLinkURL();
		void				FinishRun();
		void				InitSpecialCharHandlers();
		void				InitHandlers();
		void				FinishFontInfo();
		void				FinishColorInfo();
		void				ConsolidateText();
		void				SetATSUStyleForRun( const RTFStyleRun& inRun,
											CFURLRef inLinkURL );
		void				InsertStyleRun( const RTFStyleRun& inRun,
											TXNObject ioTextOb,
											TXNOffset& ioInsertLoc );
		
		// control word handlers
		void				Do_mac( RTFParam inParam );
		void				Do_ansicpg( RTFParam inParam );
		void				Do_fonttbl( RTFParam inParam );
		void				Do_colortbl( RTFParam inParam );
		void				Do_stylesheet( RTFParam inParam );
		void				Do_i( RTFParam inParam );	// italic
		void				Do_b( RTFParam inParam );	// bold
		void				Do_ul( RTFParam inParam );	// underline
		void				Do_uc( RTFParam inParam );	// unicode subs count
		void				Do_u( RTFParam inParam );	// unicode character
		void				Do_plain( RTFParam inParam );
		void				Do_f( RTFParam inParam );	// font
		void				Do_fs( RTFParam inParam );	// font size
		void				Do_cf( RTFParam inParam );	// foreground color
		void				Do_red( RTFParam inParam );
		void				Do_green( RTFParam inParam );
		void				Do_blue( RTFParam inParam );
		void				Do_super( RTFParam inParam );
		void				Do_sub( RTFParam inParam );
		void				Do_nosupersub( RTFParam inParam );
		void				Do_strike( RTFParam inParam );
		void				Do_fmodern( RTFParam inParam );
		void				Do_froman( RTFParam inParam );
		void				Do_fswiss( RTFParam inParam );
		void				Do_fldinst( RTFParam inParam );
		void				Do_fldrslt( RTFParam inParam );
	
		RTFDataStream		mStream;
		int					mNestLevel;
		bool				mIsMacRoman;
		FontTable			mFontTable;
		ColorVec			mColorTable;
		RTFColor			mColor;	// used in parsing color table
		RunVec				mStyleRuns;
		CFMutableStringRefHolder	mNextRunUnicode;
		std::string			mNextRunWinANSIOrMacRoman;
		StateVec			mStates;
		HandlerMap			mHandlers;
		TokenToChar			mSpecialCharHandlers;
		ATSUStyle			mWorkStyle;
		EFontType			mFontType;
		CFStringRefHolder	mLinkURL;
	};
}

bool	operator==( const RTFColor& inOne, const RTFColor& inTwo )
{
	return (inOne.red == inTwo.red) and
			(inOne.green == inTwo.green) and
			(inOne.blue == inTwo.blue);
}

CRTFParser::CRTFParser( CFDataRef inData )
	: mStream( inData )
	, mNestLevel( 0 )
	, mIsMacRoman( false )
	, mColor( kBlackColor )
	, mNextRunUnicode( CFStringCreateMutable( NULL, 0 ) )
	, mFontType( kFontType_Unknown )
{
	RTFStateFrame	startState = {
		kRTFDestination_Main,
		1,
		{
			0,
			0,
			24,
			normal,
			kBaselineShift_None,
			false
		}
	};
	mStates.push_back( startState );
	
	InitSpecialCharHandlers();
	InitHandlers();
	
	OSStatus	err = ATSUCreateStyle( &mWorkStyle );
	if (err != noErr)
	{
		throw std::exception();
	}
}

CRTFParser::~CRTFParser()
{
	ATSUDisposeStyle( mWorkStyle );
}

void	CRTFParser::InitSpecialCharHandlers()
{
	struct	SpecialDef
	{
		const char*	token;
		CharPair	chars;
	};
	
	SpecialDef	theSpecials[] =
	{
		{ "tab", { '\t', '\t' } },
		{ "par", { '\n', '\n' } },
		{ "line", { '\n', '\n' } },
		{ "lquote", { 0xD4, 0x91 } },
		{ "rquote", { 0xD5, 0x92 } },
		{ "ldblquote", { 0xD2, 0x93 } },
		{ "rdblquote", { 0xD3, 0x94 } },
		{ "bullet", { 0xA5, 0x95 } },
		{ "endash", { 0xD1, 0x96 } },
		{ "emdash", { 0xD0, 0x97 } }
	};
	
	for (int i = 0; i < sizeof(theSpecials)/sizeof(theSpecials[0]); ++i)
	{
		mSpecialCharHandlers[ theSpecials[i].token ] = theSpecials[i].chars;
	}
}

void	CRTFParser::InitHandlers()
{
	CSDef	theHandlers[] =
	{
		{ "mac", &CRTFParser::Do_mac },
		{ "ansicpg", &CRTFParser::Do_ansicpg },
		{ "fonttbl", &CRTFParser::Do_fonttbl },
		{ "colortbl", &CRTFParser::Do_colortbl },
		{ "red", &CRTFParser::Do_red },
		{ "green", &CRTFParser::Do_green },
		{ "blue", &CRTFParser::Do_blue },
		{ "stylesheet", &CRTFParser::Do_stylesheet },
		{ "fldinst", &CRTFParser::Do_fldinst },
		{ "fldrslt", &CRTFParser::Do_fldrslt },
		{ "i", &CRTFParser::Do_i },
		{ "b", &CRTFParser::Do_b },
		{ "ul", &CRTFParser::Do_ul },
		{ "uc", &CRTFParser::Do_uc },
		{ "u", &CRTFParser::Do_u },
		{ "plain", &CRTFParser::Do_plain },
		{ "f", &CRTFParser::Do_f },
		{ "fs", &CRTFParser::Do_fs },
		{ "cf", &CRTFParser::Do_cf },
		{ "super", &CRTFParser::Do_super },
		{ "sub", &CRTFParser::Do_sub },
		{ "nosupersub", &CRTFParser::Do_nosupersub },
		{ "strike", &CRTFParser::Do_strike },
		{ "fmodern", &CRTFParser::Do_fmodern },
		{ "froman", &CRTFParser::Do_froman },
		{ "fswiss", &CRTFParser::Do_fswiss }
	};

	for (int i = 0; i < sizeof(theHandlers)/sizeof(theHandlers[0]); ++i)
	{
		mHandlers[ theHandlers[i].token ] = theHandlers[i].handler;
	}
}

/*!
	@function	SetDestination
	@abstract	This is called when a control word is making us enter a new RTF
				destination.
	@param		inDest		The new destinaton.
*/
void	CRTFParser::SetDestination( ERTFDestination inDest )
{
	FinishRun();
	
	mStates.back().mDestination = inDest;
}

/*!
	@function	LookForLinkURL
	@abstract	
*/
void	CRTFParser::LookForLinkURL()
{
	CFIndex	dataLen = CFStringGetLength( mNextRunUnicode.get() );
	
	if (CFStringGetCharacterAtIndex( mNextRunUnicode.get(), dataLen-1 ) == '"')
	{
		CFRange	foundRange = CFStringFind( mNextRunUnicode.get(),
			CFSTR("HYPERLINK \""), 0 );
		if ( (foundRange.location == 0) and (foundRange.length + 1 < dataLen) )
		{
			CFStringRefHolder	theURL( CFStringCreateWithSubstring( NULL,
				mNextRunUnicode.get(),
				CFRangeMake( foundRange.length,
					dataLen - foundRange.length - 1) ) );
			mLinkURL = theURL;
		}
	}
}

/*!
	@function	ConsolidateText
	@abstract	Convert accumulated text to unicode.
	@discussion	Current text is held in two places: Unicode text in the
				mNextRunUnicode member, and following characters in WinANSI or
				MacRoman encoding in the mNextRunWinANSIOrMacRoman member.
				This method converts the mNextRunWinANSIOrMacRoman to Unicode
				and appends it to mNextRunUnicode.
*/
void	CRTFParser::ConsolidateText()
{
	if (not mNextRunWinANSIOrMacRoman.empty())
	{
		CFStringRefHolder	newText( CFStringCreateWithBytes( NULL,
			reinterpret_cast<const UInt8*>( mNextRunWinANSIOrMacRoman.data() ),
			mNextRunWinANSIOrMacRoman.size(),
			mIsMacRoman? kCFStringEncodingMacRoman :
				kCFStringEncodingWindowsLatin1,
			false ) );
		if (newText.get() != NULL)
		{
			CFStringAppend( mNextRunUnicode.get(), newText.get() );
		}
		mNextRunWinANSIOrMacRoman.clear();
	}
}

/*!
	@function		FinishRun
	@abstract		If any text has been accumulated, make a style run.
*/
void	CRTFParser::FinishRun()
{
	ConsolidateText();
	
	CFIndex	theLen = CFStringGetLength( mNextRunUnicode.get() );

	if (theLen > 0)
	{
		ERTFDestination curDest = mStates.back().mDestination;
		
		if (curDest == kRTFDestination_Main)
		{
			RTFStyleRun	theRun;
			theRun.mAtts = mStates.back().mAtts;
			CFStringRefHolder	theText( CFStringCreateCopy( NULL,
				mNextRunUnicode.get() ) );
			theRun.mText = theText;
			
			mStyleRuns.push_back( theRun );
		}
		else if (curDest == kRTFDestination_FieldResult)
		{
			RTFStyleRun	theRun;
			theRun.mAtts = mStates.back().mAtts;
			CFStringRefHolder	theText( CFStringCreateCopy( NULL,
				mNextRunUnicode.get() ) );
			theRun.mText = theText;
			theRun.mLinkURL = mLinkURL;
			
			mStyleRuns.push_back( theRun );
		}
		else if (curDest == kRTFDestination_FieldInstructions)
		{
			LookForLinkURL();
		}
	
		CFStringDelete( mNextRunUnicode.get(), CFRangeMake( 0, theLen ) );
	}
}

void	CRTFParser::FinishFontInfo()
{
	FontEntry	curFont = {
		mNextRunWinANSIOrMacRoman,
		mFontType
	};
	mFontTable[ mStates.back().mAtts.mFontNumber ] = curFont;
}

void	CRTFParser::FinishColorInfo()
{
	mColorTable.push_back( mColor );
	mColor = kBlackColor;
}

void	CRTFParser::DoBeginGroup()
{
	RTFStateFrame	topState = mStates.back();
	mStates.push_back( topState );
}

void	CRTFParser::DoEndGroup()
{
	FinishRun();
	if (mStates.size() <= 1)	// the stack should never become empty
	{
		throw std::exception();
	}
	mStates.pop_back();
}

/*!
	@function	DoControlWord
	@abstract	Handle a control word, often by dispatching to a more specific
				handler.
	@param		inToken		A control word, minus the initial backslash.
	@param		inParam		The parameter, if any.
*/
void	CRTFParser::DoControlWord( const std::string& inToken,
									RTFParam inParam )
{
	TokenToChar::iterator foundSpecial = mSpecialCharHandlers.find( inToken );
	if (foundSpecial != mSpecialCharHandlers.end())
	{
		const CharPair&	thePair( foundSpecial->second );
		DoPlainCharacter( mIsMacRoman? thePair.macChar : thePair.winChar );
	}
	else
	{
		HandlerMap::iterator foundWord = mHandlers.find( inToken );
		if (foundWord != mHandlers.end())
		{
			(this->*foundWord->second)( inParam );
		}
	}
}

void	CRTFParser::Do_mac( RTFParam inParam )
{
#pragma unused( inParam )
	mIsMacRoman = true;
}

void	CRTFParser::Do_ansicpg( RTFParam inParam )
{
	if (inParam == 10000)
	{
		mIsMacRoman = true;
	}
}

void	CRTFParser::Do_fonttbl( RTFParam inParam )
{
#pragma unused( inParam )
	SetDestination( kRTFDestination_FontTable );
}

void	CRTFParser::Do_colortbl( RTFParam inParam )
{
#pragma unused( inParam )
	SetDestination( kRTFDestination_ColorTable );
	mColor = kBlackColor;
}

void	CRTFParser::Do_fldinst( RTFParam inParam )
{
#pragma unused( inParam )
	SetDestination( kRTFDestination_FieldInstructions );
	mLinkURL = CFStringRefHolder();
}

void	CRTFParser::Do_fldrslt( RTFParam inParam )
{
#pragma unused( inParam )
	SetDestination( kRTFDestination_FieldResult );
}

void	CRTFParser::Do_red( RTFParam inParam )
{
	if (inParam == kAbsentRTFParam)
	{
		throw std::exception();
	}
	mColor.red = inParam;
}

void	CRTFParser::Do_green( RTFParam inParam )
{
	if (inParam == kAbsentRTFParam)
	{
		throw std::exception();
	}
	mColor.green = inParam;
}

void	CRTFParser::Do_blue( RTFParam inParam )
{
	if (inParam == kAbsentRTFParam)
	{
		throw std::exception();
	}
	mColor.blue = inParam;
}

void	CRTFParser::Do_stylesheet( RTFParam inParam )
{
#pragma unused( inParam )
	SetDestination( kRTFDestination_StyleSheets );
}

void	CRTFParser::Do_i( RTFParam inParam )	// italic
{
	FinishRun();
	if (inParam == 0)
	{
		mStates.back().mAtts.mStyle &= ~italic;
	}
	else
	{
		mStates.back().mAtts.mStyle |= italic;
	}
}

void	CRTFParser::Do_b( RTFParam inParam )	// bold
{
	FinishRun();
	if (inParam == 0)
	{
		mStates.back().mAtts.mStyle &= ~bold;
	}
	else
	{
		mStates.back().mAtts.mStyle |= bold;
	}
}

void	CRTFParser::Do_ul( RTFParam inParam )	// underline
{
	FinishRun();
	if (inParam == 0)
	{
		mStates.back().mAtts.mStyle &= ~underline;
	}
	else
	{
		mStates.back().mAtts.mStyle |= underline;
	}
}

void	CRTFParser::Do_super( RTFParam inParam )
{
#pragma unused( inParam )
	FinishRun();
	mStates.back().mAtts.mShift = kBaselineShift_Super;
}

void	CRTFParser::Do_sub( RTFParam inParam )
{
#pragma unused( inParam )
	FinishRun();
	mStates.back().mAtts.mShift = kBaselineShift_Sub;
}

void	CRTFParser::Do_nosupersub( RTFParam inParam )
{
#pragma unused( inParam )
	FinishRun();
	mStates.back().mAtts.mShift = kBaselineShift_None;
}

void	CRTFParser::Do_strike( RTFParam inParam )
{
	FinishRun();
	mStates.back().mAtts.mStrikethrough = (inParam != 0);
}

void	CRTFParser::Do_uc( RTFParam inParam )	// unicode subs count
{
	if (inParam == kAbsentRTFParam)
	{
		throw std::exception();
	}
	mStates.back().mUnicodeAltBytes = inParam;
}

void	CRTFParser::Do_u( RTFParam inParam )	// unicode character
{
	if (inParam == kAbsentRTFParam)
	{
		throw std::exception();
	}
	ConsolidateText();
	UniChar	theChar = inParam;
	CFStringAppendCharacters( mNextRunUnicode.get(), &theChar, 1 );
	
	int	bytesToEat = mStates.back().mUnicodeAltBytes;
	std::string	dummyToken;
	RTFParam	dummyParam;
	for (int i = 0; i < bytesToEat; ++i)
	{
		mStream.GetToken( dummyToken, dummyParam );
	}
}

void	CRTFParser::Do_plain( RTFParam inParam )
{
#pragma unused( inParam )
	FinishRun();
	mStates.back().mAtts.mStyle = normal;
	mStates.back().mAtts.mShift = kBaselineShift_None;
	mStates.back().mAtts.mStrikethrough = false;
	mStates.back().mAtts.mFontSizeHalfPoints = 24;
}

void	CRTFParser::Do_f( RTFParam inParam )	// font
{
	if (inParam == kAbsentRTFParam)
	{
		throw std::exception();
	}
	FinishRun();
	mStates.back().mAtts.mFontNumber = inParam;
	
	if (mStates.back().mDestination == kRTFDestination_FontTable)
	{
		mFontType = kFontType_Unknown;
	}
}

void	CRTFParser::Do_fmodern( RTFParam inParam )
{
#pragma unused( inParam )
	mFontType = kFontType_Modern;
}

void	CRTFParser::Do_froman( RTFParam inParam )
{
#pragma unused( inParam )
	mFontType = kFontType_Roman;
}

void	CRTFParser::Do_fswiss( RTFParam inParam )
{
#pragma unused( inParam )
	mFontType = kFontType_Swiss;
}


void	CRTFParser::Do_fs( RTFParam inParam )	// font size
{
	if (inParam == kAbsentRTFParam)
	{
		throw std::exception();
	}
	FinishRun();
	mStates.back().mAtts.mFontSizeHalfPoints = inParam;
}

void	CRTFParser::Do_cf( RTFParam inParam )	// foreground color
{
	if (inParam == kAbsentRTFParam)
	{
		throw std::exception();
	}
	FinishRun();
	mStates.back().mAtts.mColorIndex = inParam;
}

/*!
	@function	DoControlSymbol
	@abstract	Handle a control symbol.
	@param		inToken		The letter part of the control symbol.
	@param		inParam		The parameter, if any.
*/
void	CRTFParser::DoControlSymbol( char inToken,
									RTFParam inParam )
{
	switch (inToken)
	{
		case '\r':
		case '\n':
			DoPlainCharacter( '\n' );
			break;
		
		case '*':
			SetDestination( kRTFDestination_Unknown );
			break;
		
		case '-':	// optional hyphen, ignore
			break;
		
		case '_':	// nonbreaking hyphen
			DoPlainCharacter( '-' );
			break;
		
		case '\'':	// hex escape
			if (inParam != kAbsentRTFParam)
			{
				DoPlainCharacter( static_cast<char>( inParam ) );
			}
			break;
		
		default:	// e.g., escaped \, {, }
			DoPlainCharacter( inToken );
			break;
	}
}

/*!
	@function	DoPlainCharacter
	@abstract	Handle a plain character obtained from the RTF stream.
				Normally, we just append it a variable holding non-Unicode text,
				but the semicolon character is treated specially within font and
				color tables.
	@param		inChar		A character.
*/
void	CRTFParser::DoPlainCharacter( char inChar )
{
	if (inChar == ';')
	{
		switch (mStates.back().mDestination)
		{
			case kRTFDestination_FontTable:
				FinishFontInfo();
				break;
			
			case kRTFDestination_ColorTable:
				FinishColorInfo();
				break;
			
			default:
				mNextRunWinANSIOrMacRoman += inChar;
				break;
		}
	}
	else
	{
		mNextRunWinANSIOrMacRoman += inChar;
	}
}

/*!
	@function	ParseText
	@abstract	Convert the RTF data into style runs.
	@result		True if parsing was successful.
*/
bool	CRTFParser::ParseText()
{
	bool	didParse = false;
	RTFDataStream::ETokenType	tokenType;
	std::string	theToken;
	RTFParam	theParam;
	
	try
	{
		// The first 2 tokens should always be an open brace and \rtf1.
		tokenType = mStream.GetToken( theToken, theParam );
		if (tokenType != RTFDataStream::kTokenType_BeginGroup)
		{
			throw std::exception();
		}
		tokenType = mStream.GetToken( theToken, theParam );
		if ( (tokenType != RTFDataStream::kTokenType_ControlWord) or
			(theToken != "rtf") or (theParam != 1) )
		{
			throw std::exception();
		}
		mStream.Rewind();
		
		while ( (tokenType = mStream.GetToken( theToken, theParam )) !=
			RTFDataStream::kTokenType_EndOfData )
		{
			switch (tokenType)
			{
				case RTFDataStream::kTokenType_BeginGroup:
					DoBeginGroup();
					break;
					
				case RTFDataStream::kTokenType_EndGroup:
					DoEndGroup();
					break;
					
				case RTFDataStream::kTokenType_ControlWord:
					DoControlWord( theToken, theParam );
					break;
					
				case RTFDataStream::kTokenType_ControlSymbol:
					DoControlSymbol( theToken[0], theParam );
					break;
					
				case RTFDataStream::kTokenType_PlainCharacter:
					DoPlainCharacter( theToken[0] );
					break;
			}
		}
		
		didParse = true;
	}
	catch (...)
	{
	}
	
	return didParse;
}

/*!
	@function	LoadText
	@abstract	Put the contents of the style runs into the MLTE text object.
	@param		ioTextOb		The text object to be modified.
	@param		inStartOffset	Starting offset for text to be replaced.
	@param		inEndOffset		Ending offset for text to be replaced.
*/
void	CRTFParser::LoadText( TXNObject ioTextOb,
							TXNOffset inStartOffset,
							TXNOffset inEndOffset )
{
	// Clear the interval to replace.
	TXNSetData( ioTextOb, kTXNUnicodeTextData, NULL, 0,
		inStartOffset, inEndOffset );
	
	for (RunVec::iterator i = mStyleRuns.begin(); i != mStyleRuns.end(); ++i)
	{
		InsertStyleRun( *i, ioTextOb, inStartOffset );
	}
}

/*!
	@function	InsertPlainText
	@abstract	Insert a run of plain text into the MLTE text object.
	@param		inText			A text string.
	@param		ioTextOb		An MLTE text object.
	@param		inStartOffset	Offset at which to insert.
*/
static void InsertPlainText( CFStringRef inText, TXNObject ioTextOb,
							TXNOffset inStartOffset )
{
	// Get text as unicode, and insert it.
	CFIndex	newCharCount = CFStringGetLength( inText );
	std::vector<UniChar> charBuf( newCharCount );
	CFStringGetCharacters( inText, CFRangeMake( 0, newCharCount ),
		&charBuf[0] );
	TXNSetData( ioTextOb, kTXNUnicodeTextData, &charBuf[0],
		newCharCount * sizeof(UniChar),
		inStartOffset, inStartOffset );
}

/*!
	@function	GetFontIDFromName
	@abstract	Given a font name, assumed to be a PostScript name, find the
				ATSU font ID.
	@param		inName		A font name.
	@param		inEncoding	A character encoding for the name.
	@result		A font ID, or kInvalidFont.
*/
static ATSUFontID	GetFontIDFromName( const std::string& inName,
										CFStringEncoding inEncoding )
{
	ATSUFontID	fontID = kInvalidFont;
	
	CFStringRefHolder	fontNameCF( CFStringCreateWithBytes( NULL,
		reinterpret_cast<const UInt8*>( inName.data() ), inName.size(),
		inEncoding, false ) );
	if ( fontNameCF.get() != NULL )
	{
		ATSFontRef	fontRef = ATSFontFindFromPostScriptName( fontNameCF.get(),
			kATSOptionFlagsDefault );
		
		if (fontRef == kATSFontRefUnspecified)
		{
			// Although CreateRTFFromMLTE uses PostScript names, and I think
			// Cocoa does too, RTF coming from a different source might use a
			// different kind of name.
			fontRef = ATSFontFindFromName( fontNameCF.get(),
				kATSOptionFlagsDefault );
		}
		
		if (fontRef != kATSFontRefUnspecified)
		{
			fontID = FMGetFontFromATSFontRef( fontRef );
		}
	}
	return fontID;
}

/*!
	@function	GetFontIDFromFontEntry
	@abstract	Given a font table entry (font name and faily type), find an
				ATSU font ID.
	@param		inFont		A font table entry.
	@param		inEncoding	Encoding of the font name.
	@result		A font ID.
*/
static ATSUFontID	GetFontIDFromFontEntry( const FontEntry& inFont,
										CFStringEncoding inEncoding )
{
	ATSUFontID	fontID = GetFontIDFromName( inFont.mName, inEncoding );
	
	if (fontID == kInvalidFont) // we don't have that exact font
	{
		switch (inFont.mType)
		{
			case kFontType_Roman:
				fontID = GetFontIDFromName( "Times-Roman", smRoman );
				break;
				
			case kFontType_Modern:
				fontID = GetFontIDFromName( "Courier", smRoman );
				break;
				
			case kFontType_Swiss:
				fontID = GetFontIDFromName( "Helvetica", smRoman );
				break;
				
			default:
				fontID = GetFontIDFromName( "LucidaGrande", smRoman );
				break;
		}
	}
	
	if (fontID == kInvalidFont)	// quite unlikely
	{
		fontID = GetFontIDFromName( "LucidaGrande", smRoman );
	}
	
	return fontID;
}

/*!
	@function	RTFtoRGBColor
	@abstract	Convert an RTFColor to an RGBColor.
	@param		inColor		An RTF color.
	@result		An RGB color.
*/
static RGBColor	RTFtoRGBColor( const RTFColor& inColor )
{
	RGBColor	theQDColor;
	theQDColor.red = (((int) inColor.red) << 8) | inColor.red;
	theQDColor.green = (((int) inColor.green) << 8) | inColor.green;
	theQDColor.blue = (((int) inColor.blue) << 8) | inColor.blue;
	return theQDColor;
}

/*!
	@function	SetATSUStyleForRun
	@abstract	Apply the attributes of the given run to the mWorkStyle member.
	@param		inRun		A style run.
	@param		inLinkURL	A URL to use for a link.
*/
void	CRTFParser::SetATSUStyleForRun( const RTFStyleRun& inRun,
										CFURLRef inLinkURL )
{
	// Get attributes of new text
	ATSUFontID	fontID = GetFontIDFromFontEntry(
		mFontTable[ inRun.mAtts.mFontNumber ],
		mIsMacRoman? kCFStringEncodingMacRoman :
			kCFStringEncodingWindowsLatin1 );

	RGBColor	theQDColor = { 0, 0, 0 };
	if ( (inRun.mAtts.mColorIndex >= 0) and
		(inRun.mAtts.mColorIndex < mColorTable.size()) )
	{
		theQDColor = RTFtoRGBColor( mColorTable[ inRun.mAtts.mColorIndex ] );
	}

	Fixed	shiftDelta = 0;
	if (inRun.mAtts.mShift != kBaselineShift_None)
	{
		Fixed	fontSize = (inRun.mAtts.mFontSizeHalfPoints << 15);
		shiftDelta = fontSize / 4;
		if (inRun.mAtts.mShift == kBaselineShift_Sub)
		{
			shiftDelta = - shiftDelta;
		}
	}
	
	Boolean	theStrike = inRun.mAtts.mStrikethrough;
	Boolean	theBold = (inRun.mAtts.mStyle & bold) != 0;
	Boolean	theItalic= (inRun.mAtts.mStyle & italic) != 0;
	Boolean	theUnderline = (inRun.mAtts.mStyle & underline) != 0;
	
	// Set up ATSU style.
	ATSUClearStyle( mWorkStyle );
	ATSUAttributeTag	attTags[] = {
		kATSUFontTag,
		kATSUColorTag,
		kATSUCrossStreamShiftTag,
		kATSUStyleStrikeThroughTag,
		kATSUQDBoldfaceTag,
		kATSUQDItalicTag,
		kATSUQDUnderlineTag
	};
	ByteCount		attSizes[] = {
		sizeof(ATSUFontID),
		sizeof(RGBColor),
		sizeof(Fixed),
		sizeof(Boolean),
		sizeof(Boolean),
		sizeof(Boolean),
		sizeof(Boolean),
	};
	ATSUAttributeValuePtr	attValues[] = {
		&fontID,
		&theQDColor,
		&shiftDelta,
		&theStrike,
		&theBold,
		&theItalic,
		&theUnderline
	};
	ATSUSetAttributes( mWorkStyle, sizeof(attTags)/sizeof(attTags[0]),
		attTags, attSizes, attValues );
	
	if (inLinkURL != NULL)
	{
		// This is a hack to work around an OS bug: You can set
		// kTXNURLAttribute, but you can't get it.  However, MLTE does retain
		// the CFURLRef.  <rdar://4619809>
		attTags[0] = kATSULinkURLHack;
		attSizes[0] = sizeof(CFURLRef);
		CFURLRef	theURL = inLinkURL;
		attValues[0] = &theURL;
		ATSUSetAttributes( mWorkStyle, 1, attTags, attSizes, attValues );
	}
}


/*!
	@function	InsertStyleRun
	@abstract	Insert a run of styled text into the MLTE text.
	@param		inRun		A style run.
	@param		ioTextOb	An MLTE text object.
	@param		ioInsertLoc	On input, offset at which to insert.  On output,
							receives the offset at the end of the inserted text.
*/
void	CRTFParser::InsertStyleRun( const RTFStyleRun& inRun,
									TXNObject ioTextOb,
									TXNOffset& ioInsertLoc )
{
	InsertPlainText( inRun.mText.get(), ioTextOb, ioInsertLoc );
	
	TXNOffset	endRunOff = ioInsertLoc +
		CFStringGetLength( inRun.mText.get() );

	// If there is a link URL, make a CFURL.
	CFURLRefHolder	theURL;
	if (inRun.mLinkURL.get() != NULL)
	{
		CFURLRefHolder	newURL( CFURLCreateWithString( NULL,
			inRun.mLinkURL.get(), NULL ) );
		theURL = newURL;
	}
	
	SetATSUStyleForRun( inRun, theURL.get() );
	
	TXNTypeAttributes	atsuAtts[] = {
		{ kTXNATSUIStyle, kTXNATSUIStyleSize },
		{ kATSUSizeTag, sizeof(Fixed) },
		{ kTXNURLAttribute, sizeof(CFURLRef) }
	};
	atsuAtts[0].data.dataPtr = mWorkStyle;
	atsuAtts[1].data.dataValue = (inRun.mAtts.mFontSizeHalfPoints << 15);
	atsuAtts[2].data.urlReference = theURL.get();
	// We have to set the font size separately, rather than as part of the
	// ATSUStyle, to work around a bug.  <rdar://4632224>

	// Apply the ATSUStyle to the run.
	TXNSetTypeAttributes( ioTextOb, sizeof(atsuAtts)/sizeof(atsuAtts[0]),
		atsuAtts, ioInsertLoc, endRunOff );
	
	ioInsertLoc = endRunOff;
}

/*!
	@function	LoadMLTEFromRTF
	
	@abstract	Replace an interval in an MLTE object with text from decoding
				RTF data.
	
	@param		inRTFData		RTF data for styled text.
	@param		ioTextOb		An MLTE text object.
	@param		inStartOffset	Starting offset for text to be replaced.
	@param		inEndOffset		Ending offset for text to be replaced.
*/
void	LoadMLTEFromRTF( CFDataRef inRTFData,
						TXNObject ioTextOb,
						TXNOffset inStartOffset,
						TXNOffset inEndOffset )
{
	CRTFParser	parser( inRTFData );
	if (parser.ParseText())
	{
		parser.LoadText( ioTextOb, inStartOffset, inEndOffset );
	}
}
