/*
 *  RTFDataStream.cpp
 *  MLTERTFLib
 *
 *  Created by James Walker on 7/3/06.
 *
 *  Copyright (c) 2006 James W. Walker.
 *
 *	This software is provided 'as-is', without any express or implied warranty.
 *	In no event will the authors be held liable for any damages arising from the
 *	use of this software.
 *
 *	Permission is granted to anyone to use this software for any purpose,
 *	including commercial applications, and to alter it and redistribute it
 *	freely, subject to the following restrictions:
 *
 *	1. The origin of this software must not be misrepresented; you must not
 *	claim that you wrote the original software. If you use this software in a
 *	product, an acknowledgment in the product documentation would be appreciated
 *	but is not required.
 *
 *	2. Altered source versions must be plainly marked as such, and must not be
 *	misrepresented as being the original software.
 *
 *	3. This notice may not be removed or altered from any source distribution.
 */

#include "RTFDataStream.h"

#include <sstream>

RTFDataStream::RTFDataStream( CFDataRef inData )
	: mData( reinterpret_cast<const char*>( CFDataGetBytePtr( inData ) ) )
	, mDataLen( CFDataGetLength( inData ) )
	, mCurOffset( 0 )
{
}

/*!
	@function		GetChar
	@abstract		Get the next character from the data stream.
	@param			outChar		Receives the character.
	@result			True if a character was returned, false at end of file.
*/
bool	RTFDataStream::GetChar( char& outChar )
{
	bool	gotChar = false;
	if (mCurOffset < mDataLen)
	{
		outChar = mData[ mCurOffset ];
		++mCurOffset;
		gotChar = true;
	}
	return gotChar;
}

/*!
	@function		GetNonLinebreakChar
	@abstract		Get the next non-linebreak character from the data
					stream.
	@param			outChar		Receives the character.
	@result			True if a character was returned, false at end of file.
*/
bool	RTFDataStream::GetNonLinebreakChar( char& outChar )
{
	bool	gotChar;
	do
	{
		gotChar = GetChar( outChar );
	} while ( gotChar and ((outChar == '\r') or (outChar == '\n')) );
	
	return gotChar;
}


/*!
	@function		UnGetChar
	@abstract		Back up the data stream by one character.
*/
void	RTFDataStream::UnGetChar()
{
	if (mCurOffset > 0)
	{
		--mCurOffset;
	}
}

static bool IsASCIILetter( char inChar )
{
	return ((inChar >= 'a') and (inChar <= 'z')) or
		((inChar >= 'A') and (inChar <= 'Z'));
}


static bool IsDecimalDigit( char inChar )
{
	return ((inChar >= '0') and (inChar <= '9'));
}

static bool IsHexDigit( char inChar )
{
	return ((inChar >= '0') and (inChar <= '9')) or
		((inChar >= 'A') and (inChar <= 'F')) or
		((inChar >= 'a') and (inChar <= 'f'));
}

/*!
	@function		GetHexEscapeParam
	@abstract		Having read a \' control symbol, fetch the two hex
					digits that should follow.
	@param			outParam	Receives the parameter.
*/
void	RTFDataStream::GetHexEscapeParam( RTFParam& outParam )
{
	outParam = kAbsentRTFParam;
	std::string	paramStr;
	char	theChar;
	if (GetChar( theChar ))
	{
		if (IsHexDigit( theChar ))
		{
			paramStr += theChar;
			
			if (GetChar( theChar ))
			{
				if (IsHexDigit( theChar ))
				{
					paramStr += theChar;
				}
				else
				{
					UnGetChar();
				}
			}
		}
		else
		{
			UnGetChar();
		}
	}
	
	if (paramStr.size() == 2)
	{
		std::istringstream	iss( paramStr );
		iss >> std::hex >> outParam;
	}
}

/*!
	@function		GetControlWordRemainder
	@abstract		Having read the initial backslash in first letter of a
					control word, read the remaining letters and the
					numeric parameter if any.
	@param			outToken	Receives a token, i.e., a control
								word minus its backslash.
	@param			outParam	Receives the parameter, if any, of a control
								word.
*/
void	RTFDataStream::GetControlWordRemainder( std::string& ioToken,
							RTFParam& outParam )
{
	bool	gotChar;
	char	theChar;
	outParam = kAbsentRTFParam;
	
	// Get the remaining letters.
	while ( (gotChar = GetChar( theChar )) and IsASCIILetter( theChar ) )
	{
		ioToken += theChar;
	}
	
	// Either we reached the end of the data, or we have a hyphen or digit
	// starting a parameter, or we have some other terminating character.
	if (gotChar)
	{
		if ( (theChar == '-') or IsDecimalDigit( theChar ) )
		{
			std::string	paramStr;
			paramStr += theChar;
			
			// Look for more digits.
			while ( (gotChar = GetChar( theChar )) and
					IsDecimalDigit( theChar ) )
			{
				paramStr += theChar;
			}
			
			if ( gotChar and (theChar != ' ') )
			{
				UnGetChar();
			}
			
			if (paramStr != "-")
			{
				std::istringstream	iss( paramStr );
				iss >> std::dec >> outParam;
			}
		}
		else if (theChar != ' ') // terminating character not to be consumed
		{
			UnGetChar();
		}
	}
}

/*!
	@function		GetToken
	@abstract		Get the next token from the stream.
	@discussion		The backslash that begins a control word or control
					symbol is consumed but not reported as part of the
					token.
					
					In the case of a control word or a hex escape,
					there may be a parameter along with the token.  A space
					that ends a control word is consumed but not reported
					as part of the token or parameter.
	@param			outToken	Receives a token, i.e., either a control
								word or control symbol minus its backslash,
								or a single plain character.
	@param			outParam	Receives the parameter, if any, of a control
								word or hex-escape control symbol.
	@result			Type of token.
*/
RTFDataStream::ETokenType	RTFDataStream::GetToken( std::string& outToken,
												RTFParam& outParam )
{
	ETokenType	tokenType = kTokenType_EndOfData;
	outToken.clear();
	outParam = kAbsentRTFParam;
	char	firstChar;
	if (GetNonLinebreakChar( firstChar ))
	{
		if (firstChar == '{')
		{
			tokenType = kTokenType_BeginGroup;
		}
		else if (firstChar == '}')
		{
			tokenType = kTokenType_EndGroup;
		}
		else if (firstChar == '\\')
		{
			char	nextChar;
			if (GetChar( nextChar ))
			{
				outToken += nextChar;
				
				if (IsASCIILetter( nextChar ))
				{
					tokenType = kTokenType_ControlWord;
					GetControlWordRemainder( outToken, outParam );
				}
				else
				{
					tokenType = kTokenType_ControlSymbol;
					if (nextChar == '\'')
					{
						GetHexEscapeParam( outParam );
					}
				}
			}
		}
		else
		{
			tokenType = kTokenType_PlainCharacter;
			outToken += firstChar;
		}
	}
	return tokenType;
}

