/*
 *  RTFDataStream.h
 *  MLTERTFLib
 *
 *  Created by James Walker on 7/3/06.
 *
 *  Copyright (c) 2006 James W. Walker.
 *
 *	This software is provided 'as-is', without any express or implied warranty.
 *	In no event will the authors be held liable for any damages arising from the
 *	use of this software.
 *
 *	Permission is granted to anyone to use this software for any purpose,
 *	including commercial applications, and to alter it and redistribute it
 *	freely, subject to the following restrictions:
 *
 *	1. The origin of this software must not be misrepresented; you must not
 *	claim that you wrote the original software. If you use this software in a
 *	product, an acknowledgment in the product documentation would be appreciated
 *	but is not required.
 *
 *	2. Altered source versions must be plainly marked as such, and must not be
 *	misrepresented as being the original software.
 *
 *	3. This notice may not be removed or altered from any source distribution.
 */

#include <CoreFoundation/CoreFoundation.h>
#include <string>

/*!
	@typedef	RTFParam
	@abstract	Either a numeric parameter to a control word, or a special
				constant indicating the absence of a parameter.
*/
typedef	long	RTFParam;

const RTFParam	kAbsentRTFParam	= 0x7FFFFFFF;

/*!
	@class		RTFDataStream
	
	@abstract	Object to provide lexical scanning of RTF data for input.
*/
class RTFDataStream
{
public:
	enum ETokenType
	{
		kTokenType_EndOfData = 0,
		kTokenType_BeginGroup,
		kTokenType_EndGroup,
		kTokenType_ControlWord,
		kTokenType_ControlSymbol,
		kTokenType_PlainCharacter
	};

	/*!
		@function		RTFDataStream::RTFDataStream
		@abstract		Construct a RTFDataStream from a CFDataRef.
		@discussion		The CFDataRef is assumed to remain valid and unchanged
						during the life of the RTFDataStream object.
	*/
						RTFDataStream( CFDataRef inData );
						~RTFDataStream() {}
	
	/*!
		@function		GetToken
		@abstract		Get the next token from the stream.
		@discussion		The backslash that begins a control word or control
						symbol is consumed but not reported as part of the
						token.
						
						In the case of a control word or a hex escape,
						there may be a parameter along with the token.  A space
						that ends a control word is consumed but not reported
						as part of the token or parameter.
		@param			outToken	Receives a token, i.e., either a control
									word or control symbol minus its backslash,
									or a single plain character.
		@param			outParam	Receives the parameter, if any, of a control
									word or hex-escape control symbol.
		@result			Type of token.
	*/
	ETokenType			GetToken( std::string& outToken,
								RTFParam& outParam );

	/*!
		@function		Rewind
		@abstract		Back up the stream to the beginning.
	*/
	void				Rewind()  { mCurOffset = 0; }
	
private:
	/*!
		@function		GetChar
		@abstract		Get the next character from the data stream.
		@param			outChar		Receives the character.
		@result			True if a character was returned, false at end of file.
	*/
	bool				GetChar( char& outChar );

	/*!
		@function		GetNonLinebreakChar
		@abstract		Get the next non-linebreak character from the data
						stream.
		@param			outChar		Receives the character.
		@result			True if a character was returned, false at end of file.
	*/
	bool				GetNonLinebreakChar( char& outChar );
	
	/*!
		@function		UnGetChar
		@abstract		Back up the data stream by one character.
	*/
	void				UnGetChar();
	
	/*!
		@function		GetHexEscapeParam
		@abstract		Having read a \' control symbol, fetch the two hex
						digits that should follow.
		@param			outParam	Receives the parameter.
	*/
	void				GetHexEscapeParam( RTFParam& outParam );
	
	/*!
		@function		GetControlWordRemainder
		@abstract		Having read the initial backslash in first letter of a
						control word, read the remaining letters and the
						numeric parameter if any.
		@param			outToken	Receives a token, i.e., a control
									word minus its backslash.
		@param			outParam	Receives the parameter, if any, of a control
									word.
	*/
	void				GetControlWordRemainder( std::string& ioToken,
								RTFParam& outParam );

	const char*			mData;
	CFIndex				mDataLen;
	CFIndex				mCurOffset;
};
